#!/usr/bin/env python
from scapy.all import *
import requests
import json
import os
import stat
import socket
import argparse
import logging
import datetime
import uuid
import time
import calendar
import shutil
import urllib3
## to delete
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# argurments check
parser = argparse.ArgumentParser()
parser.add_argument("-v","--version", action='version', version='%(prog)s : version v.1.1')
parser.add_argument("-define", action='store_const', const='define', dest='funct', help='define variables', metavar='')
parser.add_argument("-pattern", action='store_const', const='pattern', dest='funct', help='research a pattern in GigaStor')
parser.add_argument("-regex", action='store_const', const='regex', dest='funct', help='research a regex in GigaStor')
parser.add_argument("-search", action='store_const', const='search', dest='funct', help='custom research in GigaStor')
parser.add_argument("--debug", help='debug level', action='store_const', const='debug', metavar='')
whattodo = parser.parse_args()
args, leftovers = parser.parse_known_args()

if args.debug is not None :
  logging.basicConfig(level=logging.DEBUG)
  logging.getLogger("requests").setLevel(logging.DEBUG)
else :
  logging.basicConfig(level=logging.INFO)
  logging.getLogger("requests").setLevel(logging.WARNING)
logger = logging.getLogger()
formatter = logging.Formatter('%(asctime)s :: %(levelname)s :: %(message)s')
handler = logging.FileHandler('gigastor.log')
logger.addHandler(handler)
handler.setFormatter(formatter)

# variables
def writevar(name,value):
  write = {name:value}
  try:
    with open('var.json', 'a') as var:
      json.dump(write, var)
      var.write(os.linesep)
      var.close()
  except IOError as e:
    logger.info("impossible to write var.json, please check permission/disk space...")
    exit(1)

def readvar(name,file):
  data = []
  with open(file, 'r') as var:
    for line in var:
      data.append(json.loads(line))
    for key in data:
      if name in key:
        return key[name]

def defvariables():
  if calendar.timegm(time.gmtime()) - calendar.timegm(time.localtime()) != 0:
    print("Please check if your system is in UTC?")
  logpath = input("Full path to write PCAP: ")
  while logpath is '':
    logpath = input("Please enter Full path to write PCAP: ")
  if logpath.endswith("/"):
    writevar("logpath",logpath)
  else:
    writevar("logpath",logpath + "/")
  writevar("baseURL","https://")
  writevar("sessionURL","/obs/api/v1/session")
  writevar("extractURL","/obs/api/v1/gs/traces")
  writevar("jobID","/obs/api/v1/jobs")
  ipAddress = input("GigaStor IP Address/DNS name: ")
  while ipAddress is '':
    ipAddress = input("Please enter GigaStor IP Address/DNS name: ")
  writevar("ipAddress",ipAddress)
  username = input("GigaStor User: ")
  while username is '':
    username = input("Please enter GigaStor User: ")
  writevar("username",username)
  password = input("GigaStor Password: ")
  while password is '':
    password = input("Please enter GigaStor Password: ")
  writevar("password",password)
  instance = input("GigaStor Instance Name: ")
  while instance is '':
    instance = input("Please enter GigaStor Instance Name: ")
  writevar("instanceName",instance)
  print("var.json created successfully, you can run ./observer.a.py -r or -p")

# http methods
def getRequest(sessionCookie, url):
  Cookie = "SessionAnalyzerId="+sessionCookie
  headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
  request = requests.get(url, headers = headers, verify = False)
  logger.debug("HTTP GET is " + str(request) + "\n")
  return request

def getPCAPNG(sessionCookie, url):
  Cookie = "SessionAnalyzerId="+sessionCookie
  headers = {'Content-Type':'application/x-pcapng','Accept':'*/*','Cookie':Cookie}
  request = requests.get(url, headers = headers, verify = False)
  logger.debug("HTTP GET is " + str(request) + "\n")
  return request

def postRequest(sessionCookie, url, data):
  Cookie = "SessionAnalyzerId="+sessionCookie
  headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
  request = requests.post(url, headers = headers, data = data, verify = False)
  logger.debug("HTTP POST is " + str(request) + "\n")
  return request

def putRequest(sessionCookie, url, data):
  Cookie = "SessionAnalyzerId="+sessionCookie
  headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
  request = requests.put(url, headers = headers, data = data, verify = False)
  logger.debug("HTTP PUT is " + str(request) + "\n")
  return request

def deleteRequest(sessionCookie, url):
  Cookie = "SessionAnalyzerId="+sessionCookie
  headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
  request = requests.delete(url, headers = headers, verify = False)
  logger.debug("HTTP DELETE is " + str(request) + "\n")
  return request

# sessions management
def login():
  url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("sessionURL","var.json")
  data = '{"username": "'+readvar("username","var.json")+'","password": "'+readvar("password","var.json")+'"}'
  logger.debug("data request is " + str(data) + "\n")
  response = postRequest('nothing', url, data)
  logger.debug("login data response is " + str(response.text) + "\n")
  if (response.ok):
    print("Login successfully")
    cookie = response.cookies['SessionAnalyzerId']
    logger.debug("SessionAnalyzerId is " + str(cookie) + "\n")
    return cookie
  else:
    logger.info("Login failed")
    logger.debug("login data response is " + str(response.text) + "\n")
    exit(1)

# main functions
def retrieve_data(sessionCookie, time_start, time_end, file_name, filter):
  url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("extractURL","var.json")
  data = '{"instanceName": "'+readvar("instanceName","var.json")+'", "traceFormat": "pcapng", "traceFileName": "'+file_name+'", "saveTo": "tempStorage", "extractionTimeRange": {"timeStart": "'+time_start+'","timeEnd": "'+time_end+'"}, "filter": "'+filter+'"}'
  logger.debug("data request is " + str(data) + "\n")
  response = postRequest(sessionCookie, url, data)
  json_response = json.loads(response.text)
  logger.debug("retrieve data response is " + str(response.text) + "\n")
  job_id = json_response['data']['job']['id']
  logger.debug("job id is " + str(job_id) + "\n")
  return job_id

def job_id_status(sessionCookie, job_id):
  url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("jobID","var.json") + "/" + job_id
  response = getRequest(sessionCookie, url)
  json_response = json.loads(response.text)
  logger.debug("retrieve data response is " + str(response.text) + "\n")
  job_status = json_response['data']['job']['progress']
  # Improvment: check if another trace is in process with QUEUED and display a message
  print("Trace extraction in progress: " + str(job_status) +"% completed  ", end="\r" )
  while job_status is not 100:
    time.sleep(2)
    response = getRequest(sessionCookie, url)
    json_response = json.loads(response.text)
    job_status = json_response['data']['job']['progress']
    print("Trace extraction in progress: " + str(job_status) +"% completed  ", end="\r" )
  print("\nTrace extraction completed")
  print("Downloading Trace..." )

def trace_location(sessionCookie, job_id):
  url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("jobID","var.json") + "/" + job_id
  response = getRequest(sessionCookie, url)
  json_response = json.loads(response.text)
  logger.debug("retrieve data response is " + str(response.text) + "\n")
  trace_location = json_response['data']['job']['location']
  return trace_location

def download_trace(sessionCookie, tracename, job_id, trace_location):
  url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + trace_location
  response = getPCAPNG(sessionCookie, url)
  with open(tracename, 'wb') as f:
    f.write(response.content)

# first pass analysis
def read_pcap(file):
  packets = rdpcap(file)
  partial_filter = []
  for packet in packets:
    if IP in packet:
      if TCP in packet:
        partial_filter.append("ippair %s[tcp/%s]<->%s[tcp/%s]" % (packet[IP].src, packet[TCP].sport, packet[IP].dst, packet[TCP].dport))
      if UDP in packet:
        partial_filter.append("ippair %s[udp/%s]<->%s[udp/%s]" % (packet[IP].src, packet[UDP].sport, packet[IP].dst, packet[UDP].dport))
  return partial_filter

def remove_duplicates(values):
  output = []
  seen = set()
  for value in values:
    if value not in seen:
      output.append(value)
      seen.add(value)
  return output

def format_filter(list_filter):
  filter = ""
  filter_items = 0
  filter_exceed = "" 
  for build_filter in list_filter:
    if filter_items < 75:
      filter += "(" + build_filter + ")" + " OR "
      filter_items += 1 
    else:
      filter_exceed = "yes"
      break 
  filter = filter[:-3]
  logger.debug("filter is " + str(filter) + "\n")
  if filter == "":
    print("The pattern filter returns no packet, aborting")
    exit(1)
  if filter_exceed == "yes" :
    print("Information: Too many matched IP Pairs, the first 75 IP Pairs will be retrieved")
  return filter

# time
def iso(time):
  return time.isoformat() + ".00Z"

# execution scheme
try:
  with open('var.json') as file:
    logger.debug("read var.json")
except IOError as e:
  if whattodo.funct is 'define':
    logger.debug("var.json doesn't exist or not accessible trying to create a fresh var.json file now...")
  else:
    logger.info("var.json doesn't exist or not accessible please check permission or run 'observer.a.py -define'")
    exit(1)
if whattodo.funct is None:
  print ("Nothing to do")
  print ("observer.a.py -h for help")
elif whattodo.funct is 'define':
  try:
    os.remove('var.json')
  except OSError:
    pass
  defvariables()

elif 'pattern' or 'regex' in whattodo.funct:
  now = datetime.datetime.utcnow().replace(second=0,microsecond=0)
  now_format = now.strftime('%Y-%m-%d %H:%M')
  less5 = now - datetime.timedelta(minutes=5)
  less5_format = less5.strftime('%Y-%m-%d %H:%M')
  start_input = input("Start time (YYYY-MM-DD HH:MM, example: %s): "% less5_format)
  try:
    valid_date = time.strptime(start_input, '%Y-%m-%d %H:%M')
  except ValueError:
    print('Invalid date! Please try again')
    start_input = input("Start time (YYYY-MM-DD HH:MM, example: %s): "% less5_format)
    try:
      valid_date = time.strptime(start_input, '%Y-%m-%d %H:%M')
    except ValueError:
      print('Invalid date! Exiting')
      exit(1)
  start = datetime.datetime.strptime(start_input + ":00",'%Y-%m-%d %H:%M:%S')
  end_input = input("End time (YYYY-MM-DD HH:MM, example: %s): "% now_format)
  try:
    valid_date = time.strptime(end_input, '%Y-%m-%d %H:%M')
  except ValueError:
    print('Invalid date! Please try again')
    end_input = input("End time (YYYY-MM-DD HH:MM, example: %s): "% now_format)
    try:
      valid_date = time.strptime(end_input, '%Y-%m-%d %H:%M')
    except ValueError:
      print('Invalid date! Exiting')
      exit(1)
  end = datetime.datetime.strptime(end_input + ":00",'%Y-%m-%d %H:%M:%S')
  if start == end:
    print("Sorry Start time and End time can't be the same")
    exit(1)
  tracename = input("Trace name (name.pcapng): ")
  if tracename.endswith(".pcapng"):
    pass
  else:
    tracename = tracename + ".pcapng"
  if 'pattern' in whattodo.funct:
    filter = "pattern " + "\\" + "\"" + input("Pattern to search (pattern): ")
    filter += "\\" + "\""
  elif 'search' in whattodo.funct:
    filter = input("Filter: ")
  elif 'regex' in whattodo.funct:
    filter = "patternRegex " + "\\" + "\"" + input("Regex to search (regex): ")
    filter += "\\" + "\""
  sessionCookie = login()
  if 'search' in whattodo.funct:
    print("Search in progress...")
    job1 = retrieve_data(sessionCookie, iso(start), iso(end), tracename, filter)
    job_id_status(sessionCookie, job1)
    where = trace_location(sessionCookie, job1)
    download_trace(sessionCookie, tracename, job1, where)
  else:
    print("First pass filter in progress...")
    job1 = retrieve_data(sessionCookie, iso(start), iso(end), tracename, filter)
    job_id_status(sessionCookie, job1)
    where = trace_location(sessionCookie, job1)
    download_trace(sessionCookie, tracename, job1, where)
    dual_pass_filter = format_filter(remove_duplicates(read_pcap(tracename)))
    tracename_final = "final_" + tracename
    print("Second pass filter in progress...")
    job2 = retrieve_data(sessionCookie, iso(start), iso(end), tracename_final, dual_pass_filter)
    job_id_status(sessionCookie, job2)
    where = trace_location(sessionCookie, job2)
    download_trace(sessionCookie, tracename_final, job2, where)
